﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Globalization;
using System.Linq;
using System.Runtime.InteropServices;
using System.Threading;

using Microsoft.SqlServer.Dts.Pipeline;
using Microsoft.SqlServer.Dts.Pipeline.Design;
using Microsoft.SqlServer.Dts.Pipeline.Wrapper;
using Microsoft.SqlServer.Dts.Runtime;
using Microsoft.SqlServer.Dts.Runtime.Design;
using Microsoft.SqlServer.Dts.Runtime.Wrapper;

using Microsoft.Ssdqs.Cleansing.Correction.Define;
using Microsoft.Ssdqs.Component.Common.Logic;
using Microsoft.Ssdqs.Component.Common.Messages;
using Microsoft.Ssdqs.Component.Common.UI.Common;
using Microsoft.Ssdqs.Component.Common.Utilities;
using Microsoft.Ssdqs.Core.Service.DataQuality;
using Microsoft.Ssdqs.Core.Service.Export;
using Microsoft.Ssdqs.Core.Service.KnowledgebaseManagement.Define;
using Microsoft.Ssdqs.Core.Service.Process;
using Microsoft.Ssdqs.Core.Service.Requests.DataService;
using Microsoft.Ssdqs.Core.TypedCollections;
using Microsoft.Ssdqs.DataService.Define;
using Microsoft.Ssdqs.Flow.DataQuality;
using Microsoft.Ssdqs.Flow.Notification;
using Microsoft.Ssdqs.Infra.Codes;
using Microsoft.Ssdqs.Infra.Messages;
using Microsoft.Ssdqs.Infra.Utilities;
using Microsoft.Ssdqs.Matching.Define;
using Microsoft.Ssdqs.Proxy.EntryPoint;


namespace oh22is.SqlServer.DQS
{
    [DtsPipelineComponent(
        DisplayName = "DQS Matching",
        Description = "SSIS DQS Matching Transformation",
        IconResource = "oh22is.SqlServer.DQS.Matching.ico",
        UITypeName = "oh22is.SqlServer.DQS.MatchingUI, oh22is.SqlServer.DQS.MatchingUI, Version=1.0.0.0, Culture=neutral, PublicKeyToken=daba57ceffa2385b",
        ComponentType = ComponentType.Transform,
        CurrentVersion = 0)]
    public class Matching : PipelineComponent
    {

        #region Private Variables

        /// <summary>
        /// COMPONENT PROPERTIES
        /// </summary>
        private IDTSCustomProperty100 IDTS_DQS_KNOWLEDGE_BASE;
        private IDTSCustomProperty100 IDTS_ENCRYPT_CONNECTION;
        private IDTSCustomProperty100 IDTS_MAPPING_VALUES;
        private IDTSCustomProperty100 IDTS_CLEANUP_DQS_PROJECTS;
        private IDTSCustomProperty100 IDTS_SCORE_VALUE;

        /// <summary>
        /// Name of the DQS Database
        /// </summary>
        private const string _databaseName = "DQS_MAIN";

        /// <summary>
        /// The current DQ Project
        /// </summary>
        private DQProject _dqsProject;

        /// <summary>
        /// The DQS Server Name
        /// </summary>
        private string _serverName;

        /// <summary>
        /// The DQS ConnectionManager
        /// </summary>
        private ConnectionManager _dqsConnection = null;

        /// <summary>
        /// 
        /// </summary>
        private DataSourceMapping _dataSourceMapping;

        /// <summary>
        /// 
        /// </summary>
        private PipelineBuffer _matchedOutputBuffer;

        /// <summary>
        /// 
        /// </summary>
        private PipelineBuffer _unmatchedOutputBuffer;

        /// <summary>
        /// 
        /// </summary>
        private int _chunkSize;

        /// <summary>
        /// 
        /// </summary>
        private long _knowledgeBaseId;

        /// <summary>
        /// 
        /// </summary>
        private string _knowledgeBaseName;

        /// <summary>
        /// 
        /// </summary>
        private int _minConfidence;

        /// <summary>
        /// 
        /// </summary>
        private NotificationSessionInfo _currentSession;

        /// <summary>
        /// 
        /// </summary>
        private int _matchedRecordCount = 0;

        /// <summary>
        /// 
        /// </summary>
        private int _unmatchedRecordCount = 0;

        /// <summary>
        /// Number of additional columns for the matched output
        /// </summary>
        private const int _additionalColumns = 10;
    
        /// <summary>
        /// Used when an Event needs to be canceled.
        /// </summary>
        private bool cancelEvent = true;

        /// <summary>
        /// An array list of all input columns
        /// </summary>
        private ArrayList _columns;

        #endregion

        #region Component Properties

        /// <summary>
        /// Called when the component is initally added to a data flow task. 
        /// Create and configure the input and outputs of the component.
        /// </summary>
        public override void ProvideComponentProperties()
        {

            // Reset the component.
            base.RemoveAllInputsOutputsAndCustomProperties();
            ComponentMetaData.RuntimeConnectionCollection.RemoveAll();

            IDTSInput100 input = ComponentMetaData.InputCollection.New();
            input.Name = "Input";
            input.HasSideEffects = true;

            // Add Custom Properties
            AddUserProperties();

            // Connection Manager
            var connectionManager = ComponentMetaData.RuntimeConnectionCollection.New();
            connectionManager.Name = "DQSConnectionManager";
            

            // Matched Output
            IDTSOutput100 matchedOutput = ComponentMetaData.OutputCollection.New();
            matchedOutput.Name = "Matched Output";
            matchedOutput.Description = "Matched output rows are directed to this output.";
            matchedOutput.SynchronousInputID = 0;

            // Unmatched Output
            IDTSOutput100 unmatchedOutput = ComponentMetaData.OutputCollection.New();
            unmatchedOutput.Name = "Unmatched Output";
            unmatchedOutput.Description = "Unmatched output rows are directed to this output.";
            unmatchedOutput.SynchronousInputID = 0;

        }

        /// <summary>
        /// Adds user properties
        /// </summary>
        private void AddUserProperties()
        {
            
            /// DQS Knowledge Base
            IDTS_DQS_KNOWLEDGE_BASE = ComponentMetaData.CustomPropertyCollection.New();
            IDTS_DQS_KNOWLEDGE_BASE.Name = "DqsKnowledgeBase";
            IDTS_DQS_KNOWLEDGE_BASE.Description = "DqsKnowledgeBase";
            IDTS_DQS_KNOWLEDGE_BASE.ExpressionType = DTSCustomPropertyExpressionType.CPET_NONE;

            /// Mapping between input columns and domains.
            /// Stored as COLUMN | DOMAIN ; COLUMN | DOMAIN 
            IDTS_MAPPING_VALUES = ComponentMetaData.CustomPropertyCollection.New();
            IDTS_MAPPING_VALUES.Name = "MappingValues";
            IDTS_MAPPING_VALUES.Description = "MappingValues";
            IDTS_MAPPING_VALUES.ExpressionType = DTSCustomPropertyExpressionType.CPET_NONE;

            /// Encrypt Connection
            IDTS_ENCRYPT_CONNECTION = ComponentMetaData.CustomPropertyCollection.New();
            IDTS_ENCRYPT_CONNECTION.Name = "EncryptConnection";
            IDTS_ENCRYPT_CONNECTION.Description = "EncryptConnection";
            IDTS_ENCRYPT_CONNECTION.ExpressionType = DTSCustomPropertyExpressionType.CPET_NOTIFY;
            IDTS_ENCRYPT_CONNECTION.TypeConverter = typeof(System.Boolean).AssemblyQualifiedName;
            IDTS_ENCRYPT_CONNECTION.Value = false;

            /// Clean up DQS Projects after execution
            /// Currently this feature is not supported
            IDTS_CLEANUP_DQS_PROJECTS = ComponentMetaData.CustomPropertyCollection.New();
            IDTS_CLEANUP_DQS_PROJECTS.Name = "CleanUpDqsProjects";
            IDTS_CLEANUP_DQS_PROJECTS.Description = "CleanUpDqsProjects";
            IDTS_CLEANUP_DQS_PROJECTS.TypeConverter = typeof(System.Boolean).AssemblyQualifiedName;
            IDTS_CLEANUP_DQS_PROJECTS.ExpressionType = DTSCustomPropertyExpressionType.CPET_NOTIFY;
            IDTS_CLEANUP_DQS_PROJECTS.Value = true;

            /// Minimum Score value for the output
            /// Duplicates below the scores will not send to the unmatched output
            IDTS_SCORE_VALUE = ComponentMetaData.CustomPropertyCollection.New();
            IDTS_SCORE_VALUE.Name = "ScoreValue";
            IDTS_SCORE_VALUE.Description = "ScoreValue";
            IDTS_SCORE_VALUE.ExpressionType = DTSCustomPropertyExpressionType.CPET_NOTIFY;

        }

        #endregion

        #region ReinitializeMetaData
        
        /// <summary>
        /// 
        /// </summary>
        public override void ReinitializeMetaData()
        {

            if (ComponentMetaData.InputCollection.Count > 0)
            {

                IDTSInput100 input = ComponentMetaData.InputCollection[0];
                IDTSVirtualInput100 virtualInput = input.GetVirtualInput();
                IDTSOutput100 matchedOutput = ComponentMetaData.OutputCollection[0];
                IDTSOutput100 unmatchedOutput = ComponentMetaData.OutputCollection[1];

                /// Reset the component
                matchedOutput.OutputColumnCollection.RemoveAll();
                matchedOutput.ExternalMetadataColumnCollection.RemoveAll();
                matchedOutput.IsSorted = false;

                unmatchedOutput.OutputColumnCollection.RemoveAll();
                unmatchedOutput.ExternalMetadataColumnCollection.RemoveAll();
                unmatchedOutput.IsSorted = false;

                try
                {

                    IDTSOutputColumn100 matchedOutputColNew;
                    IDTSOutputColumn100 unmatchedOutputColNew;

                    /// Create special DQS specific columns for the matched output
                    matchedOutputColNew = BuildColumn(matchedOutput, "RecordId", DataType.DT_I8, 0, 0, 0, 0, "");
                    CreateExternalMetaDataColumn(matchedOutput.ExternalMetadataColumnCollection, matchedOutputColNew);

                    matchedOutputColNew = BuildColumn(matchedOutput, "ClusterId", DataType.DT_I8, 0, 0, 0, 0, "");
                    CreateExternalMetaDataColumn(matchedOutput.ExternalMetadataColumnCollection, matchedOutputColNew);

                    matchedOutputColNew = BuildColumn(matchedOutput, "ClusterRecordRelationId", DataType.DT_I8, 0, 0, 0, 0, "");
                    CreateExternalMetaDataColumn(matchedOutput.ExternalMetadataColumnCollection, matchedOutputColNew);

                    matchedOutputColNew = BuildColumn(matchedOutput, "MatchingScore", DataType.DT_R8, 0, 0, 0, 0, "");
                    CreateExternalMetaDataColumn(matchedOutput.ExternalMetadataColumnCollection, matchedOutputColNew);

                    matchedOutputColNew = BuildColumn(matchedOutput, "RuleId", DataType.DT_I8, 0, 0, 0, 0, "");
                    CreateExternalMetaDataColumn(matchedOutput.ExternalMetadataColumnCollection, matchedOutputColNew);

                    matchedOutputColNew = BuildColumn(matchedOutput, "IsPivot", DataType.DT_BOOL, 0, 0, 0, 0, "");
                    CreateExternalMetaDataColumn(matchedOutput.ExternalMetadataColumnCollection, matchedOutputColNew);

                    matchedOutputColNew = BuildColumn(matchedOutput, "Status", DataType.DT_WSTR, 255, 0, 0, 0, "");
                    CreateExternalMetaDataColumn(matchedOutput.ExternalMetadataColumnCollection, matchedOutputColNew);

                    matchedOutputColNew = BuildColumn(matchedOutput, "PairId", DataType.DT_I8, 0, 0, 0, 0, "");
                    CreateExternalMetaDataColumn(matchedOutput.ExternalMetadataColumnCollection, matchedOutputColNew);

                    matchedOutputColNew = BuildColumn(matchedOutput, "SiblingId", DataType.DT_I8, 0, 0, 0, 0, "");
                    CreateExternalMetaDataColumn(matchedOutput.ExternalMetadataColumnCollection, matchedOutputColNew);

                    matchedOutputColNew = BuildColumn(matchedOutput, "PivotId", DataType.DT_I8, 0, 0, 0, 0, "");
                    CreateExternalMetaDataColumn(matchedOutput.ExternalMetadataColumnCollection, matchedOutputColNew);


                    /// Create a output column for each input column
                    /// and related external metadata columns
                    foreach (IDTSVirtualInputColumn100 column in virtualInput.VirtualInputColumnCollection)
                    {

                        matchedOutputColNew = BuildColumn(matchedOutput, column);
                        CreateExternalMetaDataColumn(matchedOutput.ExternalMetadataColumnCollection, matchedOutputColNew);

                        unmatchedOutputColNew = BuildColumn(unmatchedOutput, column);
                        CreateExternalMetaDataColumn(unmatchedOutput.ExternalMetadataColumnCollection, unmatchedOutputColNew);

                        //virtualInput.SetUsageType(column.LineageID, DTSUsageType.UT_READONLY);
                        
                    }

                }
                catch
                {
                    ComponentUtility.FireError(base.ComponentMetaData, ComponentMessage.DataCorrectionInputContainsInvalidColumns, new object[0]);
                }
            }

            base.ReinitializeMetaData();

        }

        #endregion

        #region PreExecute

        /// <summary>
        /// PreExecute is called one time per component
        /// </summary>
        public override void PreExecute()
        {

            int inputBuffer = base.ComponentMetaData.InputCollection[0].Buffer;
            int outputBuffer = base.ComponentMetaData.OutputCollection[0].Buffer;

            /// Saves the current input column in a ArraList for later data mapping
            /// Not the best place, must change it later
            IDTSVirtualInput100 vInput = ComponentMetaData.InputCollection[0].GetVirtualInput();
            _columns = new ArrayList();
            foreach (IDTSVirtualInputColumn100 vcol in vInput.VirtualInputColumnCollection)
            {
                _columns.Add(vcol.Name);
            }

            try
            {
                if (_dqsConnection != null)
                {

                    /// Initilize the DQS Proxy
                    InitializeProxy();

                    /// Defines different settings for the DQS Project
                    Dictionary<string, string> connectionStringParameters = DataQualityConnectorFinals.GetConnectionStringParameters(_dqsConnection.ConnectionString);
                    _serverName = (connectionStringParameters["ServerName"] == "." ? "localhost" : connectionStringParameters["ServerName"]);
                    
                    /// Porjectname composed of component name and a GUID
                    object[] objArray = new object[] { ComponentMetaData.Name.ToString(), Guid.NewGuid().ToString("N") }; 
                    string _projectName = String.Format(CultureInfo.InvariantCulture, "{0}_{1}", objArray);
                    string _description = "This project was created by the SSIS oh22is.DQS.Matching component.";

                    /// Set the minimal confidence score, based on the threshold slider on the UI
                    _minConfidence = Convert.ToInt32(ComponentMetaData.CustomPropertyCollection["ScoreValue"].Value.ToString());

                    /// Cretate the DQS Client session
                    _currentSession = CreateClientSession();

                    // Create the DQS porject in the current session
                    using (var knowledgeBase = ComponentUtility.GetKnowledgebaseManagementEntryPoint(ComponentMetaData, _databaseName, _currentSession.SessionId))
                    {
                        
                        _knowledgeBaseId = GetKnowledgeBaseIdByName(ComponentMetaData.CustomPropertyCollection["DqsKnowledgeBase"].Value.ToString());
                        _knowledgeBaseName = ComponentMetaData.CustomPropertyCollection["DqsKnowledgeBase"].Value.ToString();

                        var allKbs = knowledgeBase.KnowledgebaseGet();
                        Knowledgebase KB = Array.Find(allKbs.ToArray(), element => element.Name == _knowledgeBaseName);
                        if (KB == null)
                        {
                            InternalFireError("The Knowledge Base does not exist.");
                        }

                        _dqsProject = knowledgeBase.DQProjectAdd(_projectName, _description, _knowledgeBaseId);

                        _dqsProject.InWorkPhase = DQProjectInWorkPhase.MatchingDataSourceMapping;
                        _dqsProject = knowledgeBase.DQProjectOpen(_dqsProject);
                    }

                }

                /// load Chunksize from DQS
                using (DataQualityEntryPointClient dataQualityEntryPoint = ComponentUtility.GetDataQualityEntryPoint(base.ComponentMetaData, _databaseName, _currentSession.SessionId, _dqsProject.Id))
                {
                    IDictionary<string, int> strs = dataQualityEntryPoint.Parameters();
                    _chunkSize = strs["DCChunkSize"];
                }

                _dataSourceMapping = null;

            }
            catch (Exception ex)
            {
                InternalFireError(ex.ToString());
            }
        }

        #endregion

        #region PrimeOutput

        /// <summary>
        /// Called at run time for components with asynchronous 
        /// outputs to let these components add rows to the output buffers.
        /// </summary>
        /// <param name="outputs"></param>
        /// <param name="outputIDs"></param>
        /// <param name="buffers"></param>
        public override void PrimeOutput(int outputs, int[] outputIDs, PipelineBuffer[] buffers)
        {
            
            if (buffers == null) throw new ArgumentNullException("buffers");
            
            _matchedOutputBuffer = buffers[0];
            _unmatchedOutputBuffer = buffers[1];

        }

        #endregion

        #region ProcessInput

        /// <summary>
        /// This method is called repeatedly during package execution. 
        /// It is called each time the data flow task has a full buffer provided by an upstream component. 
        /// </summary>
        /// <param name="inputID">The ID of the input object of the component.</param>
        /// <param name="buffer">The input buffer.</param>
        public override void ProcessInput(int inputID, PipelineBuffer buffer)
        {
            /// As long as the buffer provides data
            if (!buffer.EndOfRowset)
            {
                try
                {
                    
                    /// Create a ReadOnlyString Collection to upload incoming data to DQS
                    List<ReadOnlyStringCollection> lRec = new List<ReadOnlyStringCollection>();

                    while (buffer.NextRow())
                    {
                        if (!buffer.EndOfRowset)
                        {
                            
                            List<string> list = new List<string>();
                            try
                            {
                                for (int i = 0; i < buffer.ColumnCount; i++)
                                {
                                    /// Add each buffer column to the list
                                    list.Add((buffer.IsNull(i) == true ? String.Empty : buffer[i].ToString()));
                                }

                                /// Add the list to theReadOnlyStringCollection
                                lRec.Add(new ReadOnlyStringCollection(list));

                            }
                            catch (Exception ex)
                            {
                                InternalFireError(lRec.Count.ToString());
                                InternalFireError(ex.ToString());
                            }
                        }

                    }

                    if (lRec.Count > 0)
                    {
                        /// Upload records to DQS
                        UploadRecords(lRec);
                        InternalFireInformation(String.Format("{0} records uploaded.", lRec.Count().ToString()));
                        lRec.Clear();
                    }

                }
                catch (Exception exception)
                {
                    /// If an error occurred, the project will be directly deleted.
                    DeleteDQSProject();
                    throw;
                }
            }
            else
            {

                /// The matching process is started
                if (StartProcess())
                {

                    /// Send information to the log
                    InternalFireInformation("DQS matching phase is beginning.");

                    #region Matching Results Summary Stats

                    /// Write the last process summary to the log
                    WriteLastProcessSummary();

                    #endregion

                    /// If the output exists, the matched data of DQS will be retrieved in a separate thread.
                    System.Threading.Thread matchedThread = new System.Threading.Thread(matchedOutput);
                    if (ComponentMetaData.OutputCollection[0].IsAttached) matchedThread.Start();

                    /// If the output exists, the unmatched data of DQS will be retrieved in a separate thread.
                    System.Threading.Thread unmatchedThread = new System.Threading.Thread(unmatchedOutput);
                    if (ComponentMetaData.OutputCollection[1].IsAttached) unmatchedThread.Start();

                    /// Check that the threads are still running                    
                    while (matchedThread.IsAlive) { }
                    while (unmatchedThread.IsAlive) { }

                    /// Last check for EndOfRowset, possibly send EndOfRowset.
                    if (_matchedOutputBuffer.EndOfRowset == false) _matchedOutputBuffer.SetEndOfRowset();
                    if (_unmatchedOutputBuffer.EndOfRowset == false) _unmatchedOutputBuffer.SetEndOfRowset();

                    /// Update the DQS project status to MatchingExport
                    /// Actually, this should give the opportunity to open the project in the DQS client.
                    /// Currently, this is not supported by the component.
                    /// The project will be deleted after execution.
                    _dqsProject.InWorkPhase = DQProjectInWorkPhase.MatchingExport;
                    UpdateDQSProject();

                }

            }

            base.ProcessInput(inputID, buffer);
        }

        #endregion

        #region Matching Process

        /// <summary>
        ///  Starts the matching process in the DQS
        /// </summary>
        /// <returns></returns>
        private bool StartProcess()
        {

            bool retValue = false;

            /// GetDataQualityProcessManagementEntryPoint
            using (var process = ProxyEntryPointFactory.GetDataQualityProcessManagementEntryPoint(_serverName, _databaseName, _currentSession.SessionId, _dqsProject.Id))
            {

                /// Starts the matching process in the DQS
                /// ProcessID is stored in a local variable
                long processId = process.ProcessStart(ProcessType.Matching, _dataSourceMapping.Id.ToString(CultureInfo.InvariantCulture) + ',' + "null" + ',' + ((byte)MatchingClusteringOption.Transitive).ToString(CultureInfo.InvariantCulture));

                /// Checks if the process is still running
                while (process.ProcessIsRunning(processId)) { }

                /// Get the current process from DQS
                /// If the status is successful, additional process information is written to the log.
                /// Otherwise process error and reason are written to the log; the execution is cancelled.
                DataQualityProcess proc = process.ProcessGet(processId);
                if (proc.Status != CoreStatus.Success)
                {
                    InternalFireError(proc.StatusError);
                    InternalFireError(proc.StatusReason);
                }
                else
                {
                    InternalFireInformation(String.Format("CreationTime: {0}", proc.CreationTime));
                    InternalFireInformation(String.Format("Start Time: {0}", proc.StartTime));
                    InternalFireInformation(String.Format("EndTime: {0}", proc.EndTime));
                    InternalFireInformation(String.Format("ElapsedTime: {0}", proc.ElapsedTime));
                    InternalFireInformation(String.Format("Results: {0}", proc.Results));
                }

                retValue = true;
            }

            return retValue;

        }

        /// <summary>
        /// 
        /// </summary>
        private void matchedOutput()
        {
            #region Matched Output

            /// GetMatchingEntryPoint
            using (var matching = ProxyEntryPointFactory.GetMatchingEntryPoint(_serverName, _databaseName, _currentSession.SessionId, _dqsProject.Id))
            {

                /// If duplicates were found
                if (_matchedRecordCount > 0)
                {

                    /// calculate the page size
                    int maxPageNumber = ((int)Math.Ceiling((float)_matchedRecordCount / (float)_chunkSize));
                    int pageSize = _chunkSize;
                    string sortingColumn = null;
                    bool isDescendingSorting = false;

                    MatchingResultType resultType = MatchingResultType.Matched;
                    bool allCollapsed = false;
                    Collection<long> collapsedExpandedCluster = new Collection<long>();

                    /// Retrieve the pages consecutively
                    for (int pageNumber = 0; pageNumber <= maxPageNumber; pageNumber++)
                    {

                        /// Define the page and load it with the given datasource mapping id.
                        MatchingResultsPageInfo pageInfo = new MatchingResultsPageInfo(pageNumber, pageSize, sortingColumn, isDescendingSorting, _minConfidence, resultType, allCollapsed, collapsedExpandedCluster);
                        MatchingResultsPage resultsMatched = matching.MatchingResultsGetPage(pageInfo, _dataSourceMapping.Id);

                        List<ColumnNameAndType> listColumnNameAndTypes = resultsMatched.DataColumnNamesAndTypes.ToList();

                        /// Write the records to the output buffer
                        foreach (MatchingResultRecord record in resultsMatched.Records)
                        {

                            int k = 0;

                            _matchedOutputBuffer.AddRow();
                            _matchedOutputBuffer.SetInt64(k++, Convert.ToInt64(record.RecordId));
                            _matchedOutputBuffer.SetInt64(k++, Convert.ToInt64(record.ClusterId));
                            _matchedOutputBuffer.SetInt64(k++, Convert.ToInt64(record.ClusterRecordRelationId));
                            _matchedOutputBuffer.SetDouble(k++, Convert.ToDouble(record.MatchingScore));
                            _matchedOutputBuffer.SetInt64(k++, Convert.ToInt64(record.RuleId));
                            _matchedOutputBuffer.SetBoolean(k++, Convert.ToBoolean(record.IsPivot));
                            _matchedOutputBuffer.SetString(k++, record.Status.ToString());
                            _matchedOutputBuffer.SetInt64(k++, Convert.ToInt64(record.PairId));
                            _matchedOutputBuffer.SetInt64(k++, Convert.ToInt64(record.SiblingId));
                            _matchedOutputBuffer.SetInt64(k++, Convert.ToInt64(record.PivotId));

                            int i = 0;
                            foreach (var values in record.DataValues)
                            {
                                try
                                {
                                    string type = listColumnNameAndTypes[i].Type;
                                    
                                    AddValueToBuffer(_matchedOutputBuffer, 
                                        GetOutputColumn(0, listColumnNameAndTypes[i].Name), 
                                        values, 
                                        GetOutputDataType(0, listColumnNameAndTypes[i].Name));
                                    
                                    i++;
                                }
                                catch (Exception ex)
                                {
                                    _matchedOutputBuffer[i] = null;
                                    i++;
                                    InternaleFireWarning(ex.ToString());
                                }
                            }
                        }

                        /// write a progress to thee log
                        InternalFireProgress("Matched records are written.", (100 / maxPageNumber * pageNumber));

                    }
                    _matchedOutputBuffer.SetEndOfRowset();
                }
            }

            #endregion
        }

        /// <summary>
        /// 
        /// </summary>
        private void unmatchedOutput()
        {
            #region Unmatched Output

            using (var matching = ProxyEntryPointFactory.GetMatchingEntryPoint(_serverName, _databaseName, _currentSession.SessionId, _dqsProject.Id))
            {

                /// If unmatched records exist
                if (_unmatchedRecordCount > 0)
                {

                    /// calculate the page size
                    int maxPageNumber = ((int)Math.Ceiling((float)_unmatchedRecordCount / (float)_chunkSize));
                    int pageSize = _chunkSize;
                    string sortingColumn = null;
                    bool isDescendingSorting = false;

                    MatchingResultType resultType = MatchingResultType.Unmatched;
                    bool allCollapsed = false;
                    Collection<long> collapsedExpandedCluster = new Collection<long>();

                    /// Retrieve the pages consecutively
                    for (int pageNumber = 0; pageNumber <= maxPageNumber; pageNumber++)
                    {

                        /// Define the page and load it with the given datasource mapping id.
                        MatchingResultsPageInfo unmatchedPageInfo = new MatchingResultsPageInfo(pageNumber, pageSize, sortingColumn, isDescendingSorting, _minConfidence, resultType, allCollapsed, collapsedExpandedCluster);
                        MatchingResultsPage unmatchedResultsMatched = matching.MatchingResultsGetPage(unmatchedPageInfo, _dataSourceMapping.Id);

                        List<ColumnNameAndType> listColumnNameAndTypes = unmatchedResultsMatched.DataColumnNamesAndTypes.ToList();

                        /// Write the records to the output buffer
                        foreach (MatchingResultRecord record in unmatchedResultsMatched.Records)
                        {

                            int i = 0;
                            _unmatchedOutputBuffer.AddRow();

                            foreach (var values in record.DataValues)
                            {
                                try
                                {
                                    AddValueToBuffer(_unmatchedOutputBuffer, 
                                        GetOutputColumn(1, listColumnNameAndTypes[i].Name), 
                                        values, 
                                        GetOutputDataType(1, listColumnNameAndTypes[i].Name));
                                    i++;
                                }
                                catch (Exception ex)
                                {
                                    _unmatchedOutputBuffer[i] = null;
                                }
                            }
                        }

                        /// write a progress to thee log
                        InternalFireProgress("Unmatched records are written.", (100 / maxPageNumber * pageNumber));

                    }
                    _unmatchedOutputBuffer.SetEndOfRowset();
                }
            }

            #endregion
        }

        #endregion

        #region PostExecute

        /// <summary>
        /// Posts an Information message that all is done, and calls base...
        /// </summary>
        public override void PostExecute()
        {
            try
            {
                if (ComponentMetaData.CustomPropertyCollection["CleanUpDqsProjects"].Value.ToString() == "True") DeleteDQSProject();
            }
            catch 
            {
                InternaleFireWarning("DQS project could not be deleted.");
            }
            ComponentUtility.CloseClientSession(ComponentMetaData, _databaseName, _currentSession.SessionId);
            base.PostExecute();

        }

        #endregion

        #region SSIS Functions

        #region AcquireConnections

        /// <summary>
        /// AcquireConnections is called during both component design and execution. 
        /// </summary>
        /// <param name="transaction"></param>
        public override void AcquireConnections(object transaction)
        {
            if (ComponentMetaData.RuntimeConnectionCollection["DQSConnectionManager"].ConnectionManager != null)
            {
                _dqsConnection = null;
                if (ComponentMetaData.RuntimeConnectionCollection.Count > 0)
                {
                    if (ComponentMetaData.RuntimeConnectionCollection["DQSConnectionManager"].ConnectionManager != null)
                    {
                        /// Create a DQS connection
                        _dqsConnection = Microsoft.SqlServer.Dts.Runtime.DtsConvert.GetWrapper(ComponentMetaData.RuntimeConnectionCollection["DQSConnectionManager"].ConnectionManager);
                    }
                }
            }

        }

        #endregion

        #region ReleaseConnections

        /// <summary>
        /// Called repeatedly during component design, and at the end of component execution.
        /// </summary>
        public override void ReleaseConnections()
        {
            _dqsConnection = null;
        }

        #endregion

        #region OnInputPathDetached

        /// <summary>
        /// This method is called when an IDTSPath100 object is deleted from the IDTSPathCollection100 collection. 
        /// </summary>
        /// <param name="inputID"></param>
        public override void OnInputPathDetached(int inputID)
        {
            base.OnInputPathDetached(inputID);
            ComponentMetaData.OutputCollection[0].OutputColumnCollection.RemoveAll();
            ComponentMetaData.OutputCollection[1].OutputColumnCollection.RemoveAll();
        }

        #endregion

        #region OnInputPathAttached

        /// <summary>
        /// 
        /// </summary>
        /// <param name="inputID"></param>
        public override void OnInputPathAttached(int inputID)
        {
            this.ReinitializeMetaData();
        }

        #endregion

        #region PerformUpgrade

        /// <summary>
        /// Base implementation of the PerformUpgrade method
        /// Will be supported in the next versions
        /// </summary>
        /// <param name="pipelineVersion"></param>
        public override void PerformUpgrade(int pipelineVersion)
        {
            DtsPipelineComponentAttribute componentAttribute = (DtsPipelineComponentAttribute)Attribute.GetCustomAttribute(this.GetType(), typeof(DtsPipelineComponentAttribute), false);
            int runtimeVersion = componentAttribute.CurrentVersion;
            int metadataVersion = ComponentMetaData.Version;

            if (runtimeVersion != metadataVersion)
            {
                if (metadataVersion <= 0) { }
                ComponentMetaData.Version = runtimeVersion;
            }
        }

        #endregion

        #region InsertInput

        /// <summary>
        /// 
        /// </summary>
        /// <param name="insertPlacement"></param>
        /// <param name="inputID"></param>
        /// <returns></returns>
        public override IDTSInput100 InsertInput(DTSInsertPlacement insertPlacement, int inputID)
        {
            throw new Exception("Component " + ComponentMetaData.Name + " does not allow adding inputs.");
        }

        #endregion

        #region InsertOutput

        /// <summary>
        /// 
        /// </summary>
        /// <param name="insertPlacement"></param>
        /// <param name="outputID"></param>
        /// <returns></returns>
        public override IDTSOutput100 InsertOutput(DTSInsertPlacement insertPlacement, int outputID)
        {
            throw new Exception("Component " + ComponentMetaData.Name + " does not allow adding outputs.");
        }

        #endregion

        #endregion

        #region SSIS Helper Functions

        /// <summary>
        /// Create an external meta data columns
        /// </summary>
        /// <param name="externalCollection"></param>
        /// <param name="column"></param>
        private static void CreateExternalMetaDataColumn(IDTSExternalMetadataColumnCollection100 externalCollection, IDTSOutputColumn100 column)
        {
            // For each output column create an external meta data columns.
            IDTSExternalMetadataColumn100 eColumn = externalCollection.New();
            eColumn.Name = column.Name;
            eColumn.DataType = column.DataType;
            eColumn.Precision = column.Precision;
            eColumn.Length = column.Length;
            eColumn.Scale = column.Scale;

            // wire the output column to the external metadata
            column.ExternalMetadataColumnID = eColumn.ID;
        }

        /// <summary>
        ///  Writes the results of the last execution to the log
        /// </summary>
        private void WriteLastProcessSummary()
        {
            try
            {
                /// GetMatchingEntryPoint
                /// Get the MatchingResultsGetLastProcessSummary for the current project
                /// Send all information to the Log
                using (var matching = ProxyEntryPointFactory.GetMatchingEntryPoint(_serverName, _databaseName, _currentSession.SessionId, _dqsProject.Id))
                {
                    MatchingResultsSummaryStats summaryStats = matching.MatchingResultsGetLastProcessSummary();
                    InternalFireInformation(String.Format(CultureInfo.InvariantCulture, "AverageClusterSize: {0}", summaryStats.AverageClusterSize.ToString()));
                    InternalFireInformation(String.Format(CultureInfo.InvariantCulture, "MaxClusterSize: {0}", summaryStats.MaxClusterSize.ToString()));
                    InternalFireInformation(String.Format(CultureInfo.InvariantCulture, "MinClusterSize: {0}", summaryStats.MinClusterSize.ToString()));
                    InternalFireInformation(String.Format(CultureInfo.InvariantCulture, "NumberOfClusters: {0}", summaryStats.NumberOfClusters.ToString()));
                    InternalFireInformation(String.Format(CultureInfo.InvariantCulture, "NumberOfSourceRecords: {0}", summaryStats.NumberOfSourceRecords.ToString()));

                    InternalFireInformation(String.Format(CultureInfo.InvariantCulture, "NumberOfDuplicateRecords: {0}", summaryStats.NumberOfDuplicateRecords.ToString()));
                    _matchedRecordCount = summaryStats.NumberOfDuplicateRecords;

                    InternalFireInformation(String.Format(CultureInfo.InvariantCulture, "NumberOfUnmatchedRecords: {0}", summaryStats.NumberOfUnmatchedRecords.ToString()));
                    _unmatchedRecordCount = summaryStats.NumberOfUnmatchedRecords;
                }
            }
            catch (Exception ex)
            {
                InternaleFireWarning("The process summary could not be retrieved.");
            }
        }

        /// <summary>
        /// Creates a IDTSOutputColumn100 based on the Virtual Input Column
        /// </summary>
        /// <param name="output">The output for which the column should be created</param>
        /// <param name="VirtualInputColumn">The VirtualInputColumn</param>
        /// <returns>Output Column</returns>
        private IDTSOutputColumn100 BuildColumn(IDTSOutput100 output, IDTSVirtualInputColumn100 VirtualInputColumn)
        {
            return BuildColumn(output, VirtualInputColumn.Name
                , VirtualInputColumn.DataType
                , VirtualInputColumn.Length
                , VirtualInputColumn.Precision
                , VirtualInputColumn.Scale
                , VirtualInputColumn.CodePage
                , VirtualInputColumn.Description);
        }

        /// <summary>
        /// Creates a IDTSOutputColumn100 based on the Virtual Input Column
        /// </summary>
        /// <param name="output">The output for which the column should be created</param>
        /// <param name="Name">Column Name</param>
        /// <param name="dataType">Column Data Type</param>
        /// <param name="Length">Column Length</param>
        /// <param name="Precision">Column Precision</param>
        /// <param name="Scale">Column Scale</param>
        /// <param name="CodePage">Column CodePage</param>
        /// <param name="Description">Column Description</param>
        /// <returns>Output Column</returns>
        private IDTSOutputColumn100 BuildColumn(IDTSOutput100 output, string Name, DataType dataType, int Length, int Precision, int Scale, int CodePage, string Description)
        {
            IDTSOutputColumn100 Column = output.OutputColumnCollection.New();
            Column.Name = Name;
            Column.SetDataTypeProperties(dataType,
                                       Length,
                                       Precision,
                                       Scale,
                                       CodePage);
            Column.Description = Description;
            return Column;
        }

        /// <summary>
        /// Write an Error message out to the Log.
        /// </summary>
        /// <param name="message">The error message to fire</param>
        private void InternalFireError(string message)
        {
            ComponentMetaData.FireError(0, ComponentMetaData.Name, message, string.Empty, 0, out cancelEvent);
        }

        /// <summary>
        /// Write an information out to the Log.
        /// </summary>
        /// <param name="message">The information message to fire</param>
        private void InternalFireInformation(string message)
        {
            ComponentMetaData.FireInformation(0, ComponentMetaData.Name, message, string.Empty, 0, ref cancelEvent);
        }

        /// <summary>
        /// Write a waring out to the log
        /// </summary>
        /// <param name="message">The warning message to fire</param>
        private void InternaleFireWarning(string message)
        {
            ComponentMetaData.FireWarning(0, ComponentMetaData.Name, message, string.Empty, 0);
        }

        /// <summary>
        /// Write a porgress status out to the log
        /// </summary>
        /// <param name="message">The status message to fire</param>
        /// <param name="percentComplete">The progress in percent to fire</param>
        private void InternalFireProgress(string message, int percentComplete)
        {
            ComponentMetaData.FireProgress(message, percentComplete, 0, 100, ComponentMetaData.Name, ref cancelEvent);
        }

        /// <summary>
        /// Returns the column ID based on the name
        /// </summary>
        /// <param name="BufferIndex"></param>
        /// <param name="ColumnName"></param>
        /// <returns></returns>
        private int GetOutputColumn(int BufferIndex, string ColumnName)
        {

            int retValue = 0;
            IDTSExternalMetadataColumnCollection100 exCol = ComponentMetaData.OutputCollection[BufferIndex].ExternalMetadataColumnCollection;           
            for (int i = 0; i < exCol.Count; i++)
            {
                if (exCol[i].Name == ColumnName)
                {
                    retValue = i;
                    break;
                }
            }
            return retValue;

        }

        /// <summary>
        /// Retunrs the data type of a specific column 
        /// </summary>
        /// <param name="BufferIndex"></param>
        /// <param name="ColumnName"></param>
        /// <returns></returns>
        private DataType GetOutputDataType(int BufferIndex, string ColumnName)
        {

            IDTSExternalMetadataColumnCollection100 exCol = ComponentMetaData.OutputCollection[BufferIndex].ExternalMetadataColumnCollection;
            for (int i = 0; i < exCol.Count; i++)
            {
                if (exCol[i].Name == ColumnName)
                {
                    return exCol[i].DataType;
                }
            }
            return DataType.DT_EMPTY;

        }

        /// <summary>
        /// Add a value to the output buffer
        /// </summary>
        /// <param name="outputBuffer">The output buffer</param>
        /// <param name="outputBufferIndex">The index defines the column</param>
        /// <param name="value">The value</param>
        /// <param name="dataType">The SISS DataTypefor the value</param>
        private static void AddValueToBuffer(PipelineBuffer outputBuffer, int outputBufferIndex, Object value, DataType dataType)
        {
            if (value == null || String.IsNullOrEmpty(value.ToString()))
            {
                outputBuffer.SetNull(outputBufferIndex);
            }
            else
            {
                switch (dataType)
                {
                    case DataType.DT_BOOL:
                        outputBuffer.SetBoolean(outputBufferIndex, Convert.ToBoolean(value));
                        break;
                    case DataType.DT_BYTES:
                        break;
                    case DataType.DT_CY:
                        outputBuffer.SetDecimal(outputBufferIndex, Convert.ToDecimal(value));
                        break;
                    case DataType.DT_DATE:
                        outputBuffer.SetDateTime(outputBufferIndex, Convert.ToDateTime(value));
                        break;
                    case DataType.DT_DBDATE:
                        outputBuffer[outputBufferIndex] = (DateTime)value;
                        break;
                    case DataType.DT_DBTIME:
                        break;
                    case DataType.DT_DBTIMESTAMP:
                        outputBuffer.SetDateTime(outputBufferIndex, Convert.ToDateTime(value));
                        break;
                    case DataType.DT_DECIMAL:
                        outputBuffer.SetDecimal(outputBufferIndex, Convert.ToDecimal(value));
                        break;
                    case DataType.DT_FILETIME:
                        outputBuffer[outputBufferIndex] = value;
                        break;
                    case DataType.DT_GUID:
                        outputBuffer.SetGuid(outputBufferIndex, new Guid(value.ToString()));
                        break;
                    case DataType.DT_I1:
                        outputBuffer.SetSByte(outputBufferIndex, Convert.ToSByte(value));
                        break;
                    case DataType.DT_I2:
                        outputBuffer.SetInt16(outputBufferIndex, Convert.ToInt16(value));
                        break;
                    case DataType.DT_I4:
                        outputBuffer.SetInt32(outputBufferIndex, Convert.ToInt32(value));
                        break;
                    case DataType.DT_I8:
                        outputBuffer.SetInt64(outputBufferIndex, Convert.ToInt64(value));
                        break;
                    case DataType.DT_IMAGE:
                        BlobColumn colDT_IMAGE = (BlobColumn)value;
                        if (colDT_IMAGE.IsNull)
                        {
                            outputBuffer.SetNull(outputBufferIndex);
                        }
                        else
                        {
                            outputBuffer.AddBlobData(outputBufferIndex, colDT_IMAGE.GetBlobData(0, (int)colDT_IMAGE.Length));
                        }
                        break;
                    case DataType.DT_NTEXT:
                        BlobColumn colDT_NTEXT = (BlobColumn)value;
                        if (colDT_NTEXT.IsNull)
                        {
                            outputBuffer.SetNull(outputBufferIndex);
                        }
                        else
                        {
                            outputBuffer.AddBlobData(outputBufferIndex, colDT_NTEXT.GetBlobData(0, (int)colDT_NTEXT.Length));
                        }
                        break;
                    case DataType.DT_NULL:
                        outputBuffer.SetNull(outputBufferIndex);
                        break;
                    case DataType.DT_NUMERIC:
                        outputBuffer.SetDecimal(outputBufferIndex, Convert.ToDecimal(value));
                        break;
                    case DataType.DT_R4:
                        outputBuffer.SetSingle(outputBufferIndex, Convert.ToSingle(value));
                        break;
                    case DataType.DT_R8:
                        outputBuffer.SetDouble(outputBufferIndex, Convert.ToDouble(value));
                        break;
                    case DataType.DT_STR:
                        outputBuffer.SetString(outputBufferIndex, value.ToString());
                        break;
                    case DataType.DT_TEXT:
                        outputBuffer.SetNull(outputBufferIndex);
                        break;
                    case DataType.DT_UI1:
                        outputBuffer.SetByte(outputBufferIndex, Convert.ToByte(value));
                        break;
                    case DataType.DT_UI2:
                        outputBuffer.SetUInt16(outputBufferIndex, Convert.ToUInt16(value));
                        break;
                    case DataType.DT_UI4:
                        outputBuffer.SetUInt32(outputBufferIndex, Convert.ToUInt32(value));
                        break;
                    case DataType.DT_UI8:
                        outputBuffer.SetUInt64(outputBufferIndex, Convert.ToUInt64(value));
                        break;
                    case DataType.DT_WSTR:
                        outputBuffer.SetString(outputBufferIndex, value.ToString());
                        break;
                    default:
                        outputBuffer.SetNull(outputBufferIndex);
                        break;
                }
            }
        }

        /// <summary>
        /// Returns the column ID based on the name
        /// </summary>
        /// <param name="BufferIndex"></param>
        /// <param name="ColumnName"></param>
        /// <returns></returns>
        private int GetInputColumn(string ColumnName)
        {
            int retValue = 0;

            IDTSVirtualInput100 virtualInput = ComponentMetaData.InputCollection[0].GetVirtualInput();
            
            for (int i = 0; i < virtualInput.VirtualInputColumnCollection.Count; i++)
            {
                if (virtualInput.VirtualInputColumnCollection[i].Name == ColumnName)
                {
                    retValue = i;
                    break;
                }
            }

            return retValue;
        }

        /// <summary>
        /// Returns the column DataType based on the column ID
        /// </summary>
        /// <param name="ColumnID"></param>
        /// <returns></returns>
        private DataType GetInputColumnDataType(int ColumnID)
        {
            IDTSVirtualInput100 virtualInput = ComponentMetaData.InputCollection[0].GetVirtualInput();
            return virtualInput.VirtualInputColumnCollection[ColumnID].DataType;
        }

        #endregion

        #region DQS Functions

        /// <summary>
        /// 
        /// </summary>
        private void UpdateDQSProject()
        {
            using (KnowledgebaseManagementEntryPointClient knowledgebaseManagementEntryPoint = ComponentUtility.GetKnowledgebaseManagementEntryPoint(ComponentMetaData, _databaseName, _currentSession.SessionId))
            {
                _dqsProject = knowledgebaseManagementEntryPoint.DQProjectUpdate(_dqsProject);
            }

        }

        /// <summary>
        /// Loads the data records on the DQS Server
        /// </summary>
        /// <param name="bufferRecords">Records from the current buffer</param>
        /// <returns></returns>
        private bool UploadRecords(List<ReadOnlyStringCollection> bufferRecords)
        {

            bool status = false;
            try
            {
                /// GetMetadataManagementEntryPoint
                using (var metadata = ComponentUtility.GetMetadataManagementEntryPoint(_dqsConnection, _databaseName, _currentSession.SessionId, _dqsProject.Id))
                {
                    ReadOnlyCollection<ReadOnlyStringCollection> records;
                    if (_dataSourceMapping == null) _dataSourceMapping = metadata.DataSourceMappingCreate(GetDataSourceMapping());
                    records = new ReadOnlyCollection<ReadOnlyStringCollection>(bufferRecords);
                    metadata.DataSourceMappingAddRecords(_dataSourceMapping.Id, records);
                }
            }
            catch
            {
                status = false;
                throw;
            }
            return status;

        }

        /// <summary>
        /// Deletes the current DQS Project
        /// </summary>
        private void DeleteDQSProject()
        {
            KnowledgebaseManagementEntryPointClient knowledgebaseManagementEntryPoint;

            /// GetKnowledgebaseManagementEntryPoint 
            using (knowledgebaseManagementEntryPoint = ComponentUtility.GetKnowledgebaseManagementEntryPoint(ComponentMetaData, _databaseName, _currentSession.SessionId))
            {
                if (knowledgebaseManagementEntryPoint.DQProjectGetById(_dqsProject.Id) != null) knowledgebaseManagementEntryPoint.DQProjectDelete(_dqsProject);
            }
        }

        /// <summary>
        /// Returns the DQS data type for a specified column
        /// </summary>
        /// <param name="ColumnName"></param>
        /// <returns></returns>
        private FieldType GetFieldType(string ColumnName)
        {
            FieldType retValue = FieldType.None;

            DataType ColumnDataType = GetInputColumnDataType(GetInputColumn(ColumnName));
            switch (ColumnDataType)
            {

                case DataType.DT_BOOL:
                    retValue = FieldType.String;
                    break;
                case DataType.DT_BYTES:
                    retValue = FieldType.None;
                    break;
                case DataType.DT_CY:
                    retValue = FieldType.Decimal;
                    break;
                case DataType.DT_DATE:
                    retValue = FieldType.Date;
                    break;
                case DataType.DT_DBDATE:
                    retValue = FieldType.Date;
                    break;
                case DataType.DT_DBTIME:
                    retValue = FieldType.None;
                    break;
                case DataType.DT_DBTIME2:
                    retValue = FieldType.None;
                    break;
                case DataType.DT_DBTIMESTAMP:
                    retValue = FieldType.None;
                    break;
                case DataType.DT_DBTIMESTAMP2:
                    retValue = FieldType.None;
                    break;
                case DataType.DT_DBTIMESTAMPOFFSET:
                    retValue = FieldType.None;
                    break;
                case DataType.DT_DECIMAL:
                    retValue = FieldType.Decimal;
                    break;
                case DataType.DT_EMPTY:
                    retValue = FieldType.None;
                    break;
                case DataType.DT_FILETIME:
                    retValue = FieldType.None;
                    break;
                case DataType.DT_GUID:
                    retValue = FieldType.None;
                    break;
                case DataType.DT_I1:
                    retValue = FieldType.Integer;
                    break;
                case DataType.DT_I2:
                    retValue = FieldType.Integer;
                    break;
                case DataType.DT_I4:
                    retValue = FieldType.Integer;
                    break;
                case DataType.DT_I8:
                    retValue = FieldType.Integer;
                    break;
                case DataType.DT_IMAGE:
                    retValue = FieldType.None;
                    break;
                case DataType.DT_NTEXT:
                    retValue = FieldType.None;
                    break;
                case DataType.DT_NULL:
                    retValue = FieldType.None;
                    break;
                case DataType.DT_NUMERIC:
                    retValue = FieldType.Decimal;
                    break;
                case DataType.DT_R4:
                    retValue = FieldType.Decimal;
                    break;
                case DataType.DT_R8:
                    retValue = FieldType.Decimal;
                    break;
                case DataType.DT_STR:
                    retValue = FieldType.String;
                    break;
                case DataType.DT_TEXT:
                    retValue = FieldType.None;
                    break;
                case DataType.DT_UI1:
                    retValue = FieldType.Decimal;
                    break;
                case DataType.DT_UI2:
                    retValue = FieldType.Decimal;
                    break;
                case DataType.DT_UI4:
                    retValue = FieldType.Decimal;
                    break;
                case DataType.DT_UI8:
                    retValue = FieldType.Decimal;
                    break;
                case DataType.DT_WSTR:
                    retValue = FieldType.String;
                    break;
                default:
                    retValue = FieldType.None;
                    break;

            }

            return retValue;
        }

        /// <summary>
        /// Create the data source mapping
        /// </summary>
        /// <returns></returns>
        private DataSourceMapping GetDataSourceMapping()
        {
            
            /// create DataSourceMappingField collection
            Collection<DataSourceMappingField> collFieldMappings = new Collection<DataSourceMappingField>();

            /// Column <=> Domain mapping
            string[] Mapping = ComponentMetaData.CustomPropertyCollection["MappingValues"].Value.ToString().Split(';');

            int i = 0;

            /// iterate through all input columns and create a DataSourceMappingField
            foreach (var col in _columns)
            {
                
                /// checks if the column has a mapping
                string arrayRow = Array.Find(Mapping, element => element.StartsWith(col.ToString() + "|"));
                string strDomain = null;

                if (!String.IsNullOrEmpty(arrayRow)) strDomain = arrayRow.Split('|')[1];              

                DataSourceMappingField field;

                /// if there is a mapping, domain and fieldFunctionalType are set
                if (!String.IsNullOrEmpty(strDomain))
                {
                    FieldFunctionalType fieldFunctionalType = FieldFunctionalType.Domain;
                    long domainId = GetDomainIdByName(strDomain, _knowledgeBaseName);
                    if (domainId == 0)
                    {
                        domainId = GetCompositeDomainIdByName(strDomain, _knowledgeBaseName);
                        fieldFunctionalType = FieldFunctionalType.CompositeDomain;
                    }

                    field = new DataSourceMappingField(col.ToString(), fieldFunctionalType, domainId);

                }
                else
                {
                    /// if ther isno mapping FieldFunctionalType = FieldNotMapped
                    field = new DataSourceMappingField(col.ToString(), FieldFunctionalType.FieldNotMapped, 0);
                }

                /// set the DataSourceFieldType
                field.DataSourceFieldType = GetFieldType(col.ToString());
                field.Id = i++;
                collFieldMappings.Add(field);

            }

            /// create the DataSourceMapping
            object[] objArray = new object[] { "Data Source Mapping", Guid.NewGuid() };
            DataSourceMapping dataSourceMapping = new DataSourceMapping
            {
                Name = string.Format(CultureInfo.InvariantCulture, "{0} ({1})", objArray),
                Description = "This schema is used by SSIS oh22is DQS Matching Transformation",
                MappingType = DataSourceMappingType.DeDuplicationDataSourceMapping,
                DataSourceType = DataSourceType.External,
                DataSourceMappingFields = collFieldMappings,
            };

            return dataSourceMapping;
        }

        /// <summary>
        /// check if the ConnectionManager is a DQS ConnectionManager
        /// </summary>
        /// <param name="cm"></param>
        /// <returns></returns>
        private bool isDQSConnection(ConnectionManager cm)
        {
            return cm.CreationName.StartsWith("DQS");
        }

        /// <summary>
        /// GetKnowledgeBaseIdByName
        /// </summary>
        /// <param name="knowledgeBaseName">Name of the knowledge base</param>
        /// <returns></returns>
        private long GetKnowledgeBaseIdByName(string knowledgeBaseName)
        {
            long ID = 0;

            /// GetKnowledgebaseManagementEntryPoint
            using (var knowledgeBase = ProxyEntryPointFactory.GetKnowledgebaseManagementEntryPoint(_serverName, _databaseName, _currentSession.SessionId))
            {
                var allKbs = knowledgeBase.KnowledgebaseGet();
                foreach (Knowledgebase kb in allKbs)
                {
                    if (kb.Name == knowledgeBaseName)
                    {
                        ID = kb.Id;
                        break;
                    }
                }
            }

            return ID;
        }

        /// <summary>
        /// GetDomainIdByName
        /// </summary>
        /// <param name="domainName">Name of the domain</param>
        /// <param name="knowledgeBaseName">Name of the knowledge base</param>
        /// <returns></returns>
        private long GetDomainIdByName(string domainName, string knowledgeBaseName)
        {
            long ID = 0;
            using (var metadata = ProxyEntryPointFactory.GetMetadataManagementEntryPoint(_serverName, _databaseName, _currentSession.SessionId, _knowledgeBaseId))
            {
                var domain = metadata.DomainGetByName(domainName);
                if (domain != null)
                {
                    ID = domain.Id;
                }
            }
            return ID;
        }

        /// <summary>
        /// GetCompositeDomainIdByName
        /// </summary>
        /// <param name="domainName">Name of the domain</param>
        /// <param name="knowledgeBaseName">Name of the knowledge base</param>
        /// <returns></returns>
        private long GetCompositeDomainIdByName(string domainName, string knowledgeBaseName)
        {
            long ID = 0;
            using (var metadata = ProxyEntryPointFactory.GetMetadataManagementEntryPoint(_serverName, _databaseName, _currentSession.SessionId, _knowledgeBaseId))
            {
                var domain = metadata.CompositeDomainGetByName(domainName.Replace(" (CD)", ""));
                if (domain != null)
                {
                    ID = domain.Id;
                }
            }
            return ID;
        }

        /// <summary>
        /// Creates the client session
        /// </summary>
        /// <returns></returns>
        private NotificationSessionInfo CreateClientSession()
        {
            // Get the connection string parameters
            var notificationEP = ProxyEntryPointFactory.GetNotificationEntryPoint(_serverName, _databaseName);
            var clientSession = notificationEP.NotificationSessionCreate(ClientType.SsisClient);
            return clientSession;
        }

        #endregion

        #region Validations

        /// <summary>
        /// Validates the settings of the component
        /// </summary>
        /// <returns></returns>
        public override DTSValidationStatus Validate()
        {

            DTSValidationStatus dtsValidationStatus = ComponentValidation.ValidateComponentInitialData(ComponentMetaData, _additionalColumns);
            
            if (dtsValidationStatus != DTSValidationStatus.VS_ISVALID)
            {
                return dtsValidationStatus;
            }

            if (ComponentMetaData.InputCollection.Count == 0) return DTSValidationStatus.VS_ISCORRUPT;
            if (ComponentMetaData.OutputCollection.Count == 0) return DTSValidationStatus.VS_ISCORRUPT;
          
            if (!InitializeProxy())
            {
                ComponentUtility.FireError(base.ComponentMetaData, ComponentMessage.CommonNoMappingsSpecified, new object[0]);
                return DTSValidationStatus.VS_ISBROKEN;
            }

            //CommonNoMappingsSpecified
            if (ComponentMetaData.CustomPropertyCollection["MappingValues"].Value == null)
            {
                ComponentUtility.FireError(base.ComponentMetaData, ComponentMessage.CommonNoMappingsSpecified, new object[0]);
                return DTSValidationStatus.VS_ISBROKEN;
            }

            if (ComponentValidation.DoesInputColumnsMatchDomains(ComponentMetaData) == false)
            {
                ComponentMetaData.FireError(0, ComponentMetaData.Name, "Input column mapping is not set correctly.", "", 0, out cancelEvent);
                return DTSValidationStatus.VS_ISBROKEN;
            }

            int minConfidence = Convert.ToInt32(ComponentMetaData.CustomPropertyCollection["ScoreValue"].Value.ToString());
            if (minConfidence < 80 || minConfidence > 100)
            {
                ComponentMetaData.FireError(0, ComponentMetaData.Name, "Minimal Matching Score is not set correctly.", "", 0, out cancelEvent);
                return DTSValidationStatus.VS_ISBROKEN;
            }

            if (!ComponentValidation.DoesInputColumDataTypesAreValid(ComponentMetaData.InputCollection[0].GetVirtualInput()))
            {
                ComponentMetaData.FireError(0, ComponentMetaData.Name, "An invalid input column data type was specified.", "", 0, out cancelEvent);
                return DTSValidationStatus.VS_ISBROKEN;
            }

            #region DQS related validations

            if (ComponentMetaData.CustomPropertyCollection["DqsKnowledgeBase"].Value == null)
            {
                ComponentMetaData.FireError(0, ComponentMetaData.Name, "A knowledge base is not specified.", "", 0, out cancelEvent);
                return DTSValidationStatus.VS_ISBROKEN;
            }

            /// checks the DQS server settings
            try
            {
                if (_dqsConnection != null)
                {
                    VerificationResult verificationResult = ComponentUtility.VerifyServer(_dqsConnection.ConnectionString, _databaseName);
                    if (verificationResult == VerificationResult.ServerVerified)
                    {

                        /// checks if the DQS communication is valid
                        if (!ComponentUtility.IsCommunicationValid(base.ComponentMetaData, _databaseName))
                        {
                            return DTSValidationStatus.VS_ISBROKEN;
                        }

                    }
                    switch (verificationResult)
                    {

                        case VerificationResult.ClientNotCompatible:
                            {
                                ComponentUtility.FireError(base.ComponentMetaData, ComponentMessage.ServerVersionNotCompatible, new object[0]);
                                break;
                            }
                        case VerificationResult.FrameworkVersionNotCompatible:
                            {
                                object[] objArray = new object[] { "DqsInstaller.exe -upgrade" };
                                ComponentUtility.FireError(base.ComponentMetaData, ComponentMessage.ServerFrameworkNotCompatible, objArray);
                                break;
                            }
                    }
                }
                else
                {
                    ComponentUtility.FireError(base.ComponentMetaData, ComponentMessage.DataQualityConnectionManagerNoConnectionExist, new object[0]);
                    return DTSValidationStatus.VS_ISBROKEN;
                }
            }
            catch (Exception ex)
            {
                ComponentMetaData.FireError(0, ComponentMetaData.Name, "An error occurred while retrieving the connection manager.", "", 0, out cancelEvent);
                return DTSValidationStatus.VS_ISBROKEN;
            }

            #endregion

            if (!ComponentValidation.DoesExternalMetaDataMatchOutputMetaData(base.ComponentMetaData.OutputCollection[0])) return DTSValidationStatus.VS_NEEDSNEWMETADATA;
            if (!ComponentValidation.DoesExternalMetaDataMatchOutputMetaData(base.ComponentMetaData.OutputCollection[1])) return DTSValidationStatus.VS_NEEDSNEWMETADATA;

            ComponentValidation.FixInvalidInputColumnMetaData(base.ComponentMetaData.InputCollection[0]);

            /// Finally, call the base class, which validates that the LineageID of each column in the input collection
            /// matches the LineageID of a column in the VirtualInputColumnCollection.
            return base.Validate();

        }

        #endregion

        public bool InitializeProxy()
        {
            try
            {
                ProxyInit.UseEncryption = Convert.ToBoolean(ComponentMetaData.CustomPropertyCollection["EncryptConnection"].Value);
                if (!ProxyInit.IsInit)
                {
                    ProxyInit.Init(new SsisProxyInitParameters());
                }
            }
            catch
            {
                return false;
            }
            return true;
        }

    }
}



